#include <iostream>
#include "associator.h"

#include "Eigen/Core"

namespace lidar_tracker {

double ComputeSimilarity(const Track& track,
                         const ObservationProto& observation) {
  // Predict track to observation timestamp
  const auto [predicted_state, _] =
      track.track_state.estimator.ComputePrediction(observation.timestamp());
  // L2
  const double dist =
      (predicted_state.block<2, 1>(0, 0) -
       Eigen::Vector2d(observation.lidar_observation().center_x(),
                       observation.lidar_observation().center_y()))
          .norm();

  std::cout << "dist : " << dist << std::endl;
  constexpr double kMaxDist = 2000.0;
  const double similarity = dist > kMaxDist ? 0.0 : 1.0 - dist / kMaxDist;
  return similarity;
}

std::vector<int> GreedyNearestNeighbor(
    const Eigen::MatrixXd& similarity_matrix) {
  const int rows = similarity_matrix.rows();
  const int cols = similarity_matrix.cols();
  std::vector<bool> matched_rows(rows);  // observations
  std::vector<bool> matched_cols(cols);  // tracks
  std::vector<int> match_result(rows, -1);
  while (true) {
    double max_val = -std::numeric_limits<double>::max();
    int max_val_col, max_val_row;
    for (int i = 0; i < cols; ++i) {
      if (matched_cols[i]) {
        continue;
      }
      for (int j = 0; j < rows; ++j) {
        if (matched_rows[j]) {
          continue;
        }
        if (similarity_matrix(j, i) > max_val) {
          max_val = similarity_matrix(j, i);
          max_val_col = i;
          max_val_row = j;
        }
      }
    }
    if (max_val > 0.0) {
      matched_cols[max_val_col] = true;
      matched_rows[max_val_row] = true;
      match_result[max_val_row] = max_val_col;
    } else {
      break;
    }
  }
  return match_result;
}

std::vector<int> Associator::Associate(
    const std::vector<TrackPtr>& tracks,
    const std::vector<const ObservationProto*>& observations) const {
  Eigen::MatrixXd similarity_matrix;
  const int row = observations.size();
  const int col = tracks.size();
  similarity_matrix.resize(row, col);
  similarity_matrix.setConstant(-1);

  // Build similarity matrix
  for (int i = 0; i < col; ++i) {
    const auto& track = *(tracks[i]);
    for (int j = 0; j < row; ++j) {
      const auto* observation = observations[j];
      similarity_matrix(j, i) = ComputeSimilarity(track, *observation);
    }
  }

  // Solve similarity matrix
  const auto observation_to_track = GreedyNearestNeighbor(similarity_matrix);

  return observation_to_track;
}

}  // namespace lidar_tracker
